# pylint: disable=too-many-lines, attribute-defined-outside-init, no-else-continue, no-else-break
# pylint: disable=too-many-instance-attributes, too-many-locals
from __future__ import annotations

import argparse
import os
import sys
from collections import defaultdict
from itertools import chain
from math import fabs
from math import log
from math import sqrt

import numpy as np
from sqlalchemy import and_
from sqlalchemy import bindparam
from sqlalchemy import case
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import Float
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import literal_column
from sqlalchemy import MetaData
from sqlalchemy import null
from sqlalchemy import or_
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy.databases import mysql

np.seterr(divide="raise")
np.set_printoptions(linewidth=128)


tair10_engine = None
Meta3 = None


# DON'T CHANGE THIS!!! It defines the ordering from MYSQL set type. This is the "suba_location" type
LOCNAMES = (
    "cell plate",
    "cytoskeleton",
    "cytosol",
    "endoplasmic reticulum",
    "extracellular",
    "golgi",
    "mitochondrion",
    "nucleus",
    "peroxisome",
    "plasma membrane",
    "plastid",
    "unclear",
    "vacuole",
    "endosome",
)
VERBOSE = False

# something to pretty print confusion tables (can be pasted into a reStructuredText document)


def rsttable(
    rows,
    out=sys.stdout,
    prefix="",
    colmap=None,
    otherrows=False,
    limit=False,
    header=None,
    SEP=" ",
    subheader=None,
):
    if colmap is None:
        colmap = {}
    if not header:
        header = rows[0].keys() + ([otherrows] if otherrows else [])
    else:
        header = header + ([otherrows] if otherrows else [])

    mx = [len(h) for h in header]

    def fixup(r):
        if r is None:
            return "NULL"
        return r.replace(":", r"\:") if isinstance(r, str) else str(r)

    rows = [
        [fixup(r) for r in row] + ([otherrows] if otherrows else []) for row in rows
    ]
    rows = [header] + rows

    if limit:
        rows += [[otherrows if otherrows else "..." for h in header]]

    for row in rows:
        assert len(row) == len(mx), ("header and row have different sizes", row, mx)
        mx = [max(m, len(r)) for m, r in zip(mx, row)]

    mx = [colmap.get(i, m) for i, m in enumerate(mx)]

    def padd(r, mx):
        lr = len(r)
        if lr > mx:
            return r[: mx - 3] + "..."
        return r + (" " * (mx - lr))

    rows = [[padd(r, m) for m, r in zip(mx, row)] for row in rows]
    if subheader:
        icol = 0
        sh = []
        st = []
        for ncols, h in subheader:
            mmx = mx[icol : icol + ncols]
            w = sum(mmx) + len(SEP) * (len(mmx) - 1)
            if w < len(h):
                raise ValueError('subheader "%s" too long %d' % (h, w))
            icol += ncols
            if not h:
                h = ".."
            h = padd(h, w)
            sh.append(h)
            st.append("-" * w)

        subheader = [sh, st]

    h = SEP.join("=" * m for m in mx)
    w = out.write
    w(prefix)
    w(h)
    w("\n")
    if subheader:
        for row in subheader:
            w(prefix)
            w(SEP.join(row))
            w("\n")

    w(prefix)
    w(SEP.join(rows[0]))
    w("\n")
    w(prefix)
    w(h)
    w("\n")
    for row in rows[1:]:
        w(prefix)
        w(SEP.join(row))
        w("\n")

    w(prefix)
    w(h)
    w("\n")


# pretty print sqlalchemy queries
def sql2str(s):
    from sqlalchemy.sql import visitors

    def replace_bindparam(element):
        if hasattr(element, "effective_value"):
            return text(repr(element.effective_value))
        return None

    return str(visitors.replacement_traverse(s, {}, replace_bindparam))


def Mask(locations):
    mask = 0
    for l in locations:
        idx = LOCNAMES.index(l)
        mask |= 1 << idx
    return mask


# create a SUBAcon table
def mk_consensus_table(tablename, meta3, locnames):
    cf = lambda l: list(chain.from_iterable(l))

    return Table(
        tablename,
        meta3,
        Column("locus", String(32), nullable=False, primary_key=True),
        Column("suba_location", mysql.MSSet(*(('"%s"' % s) for s in LOCNAMES))),
        *(
            cf(
                [
                    [
                        Column("call%d" % (idx + 1), String(24)),
                        Column("prob%d" % (idx + 1), Float),
                    ]
                    for idx in range(len(locnames))
                ],
            )
            + [Column(n, Float, nullable=False) for n in locnames]
        ),
        mysql_charset="latin1",
        mysql_engine="MyISAM",
    )


# known predictors
PNAMES = {
    "pred_adaboost": "AdaBoost",
    "pred_wolfpsort": "WoLF PSORT",
    "pred_yloc": "YLoc",
    "pred_atp": "ATP",
    "pred_bacello": "BaCelLo",
    "pred_chlorop": "ChloroP 1.1",
    "pred_epiloc": "EpiLoc",
    "pred_ipsort": "iPSORT",
    "pred_mitopred": "MitoPred",
    "pred_mitoprot2": "MitoProt",
    "pred_multiloc2": "MultiLoc2",
    "pred_nucleo": "Nucleo",
    "pred_plantmploc": "Plant-mPLoc",
    "pred_pclr": "PCLR v0.9",
    "pred_predotar": "Predotar v1.03",
    "pred_predsl": "PredSL",
    "pred_pprowler": "PProwler 1.2",
    "pred_pts1": "PTS1",
    "pred_slpfa": "SLPFA",
    "pred_slplocal": "SLP-Local",
    "pred_subloc": "SubLoc",
    "pred_targetp": "TargetP 1.1",
    "src_msms": "MS",
    "src_gfp": "GFP",
    "atted_0": "COEX",
    "src_ppi": "PPI",
}


SECRECTORY = frozenset(
    ["golgi", "plasma membrane", "vacuole", "endoplasmic reticulum", "extracellular"],
)

NONESET = frozenset([None])


# this a class that "wraps" a predictor table
class Predictor:
    FMAP = None
    # we use frozenset so the sets are *immutable* and we can use them as hash keys
    # e.g. {frozenset(['golgi']):frozenset(['secretory'])}
    FMAP1 = {
        frozenset(["golgi"]): frozenset(["secretory"]),
        frozenset(["plasma membrane"]): frozenset(["secretory"]),
        frozenset(["vacuole"]): frozenset(["secretory"]),
        frozenset(["endoplasmic reticulum"]): frozenset(["secretory"]),
        frozenset(["extracellular"]): frozenset(["secretory"]),
        frozenset(["cytoskeleton"]): frozenset(["cytosol"]),
    }
    FMAP2 = {
        frozenset(["cytoskeleton"]): NONESET,
        frozenset(["mitochondrion"]): NONESET,
        frozenset(["plastid"]): NONESET,
        frozenset(["nucleus"]): NONESET,
        frozenset(["cytosol"]): NONESET,
        frozenset(["peroxisome"]): NONESET,
    }
    FMAPALL = {frozenset(["cytoskeleton"]): frozenset(["cytosol"])}

    NONEFEATURE = NONESET

    def __init__(self, predictor, Cat, ignore_none, gold_std):
        classes_ = list(sorted(set(Cat.values())))
        # if ignore_none:
        #    if None in classes_: classes_.remove(None)
        self.name = predictor.name
        self.predictor_ = predictor  # an sqlalchemy table
        self.classes_ = tuple(classes_)
        self.ignore_none_ = ignore_none  # none is c feature or a 'no data'
        self.Cat_ = Cat  # map of suba locations to categories or 'classes_'
        self.gold_std_ = gold_std

    def get_features_(self, tair10):
        p = self.predictor_
        q = select([p.c.suba_location.distinct()])
        # we'll pick up nulls for rows that predictor doesn't have
        q = q.select_from(tair10.outerjoin(p, tair10.c.locus == p.c.locus))
        if self.ignore_none_:
            q = q.where(p.c.suba_location != null())

        FNone = self.NONEFEATURE
        features = {
            (frozenset(s.suba_location) if s.suba_location else FNone)
            for s in tair10_engine.execute(q).fetchall()
        }
        # e.g. set([frozenset(['mitochondrion']),...,])
        # we may map golgi to 'secretory'
        if self.FMAP:
            features = {self.FMAP.get(s, s) for s in features}

        self.features = sorted(features, key=lambda s: sorted(s)[0])

        # put none feature at end
        if self.features[0] == self.NONEFEATURE:
            self.features = self.features[1:] + [self.features[0]]

        print(
            self.__class__.__name__,
            "predictor:",
            self.name,
            "ignore none:",
            self.ignore_none_,
            "features:",
            self.feat2str(),
            file=sys.stderr,
        )

    def feat2str(self):
        return [",".join(str(s) for s in sorted(f)) for f in self.features]

    def set2vec(self, suba_location):
        featureset = suba_location
        ret = [0] * len(self.features)
        if self.FMAP:
            featureset = self.FMAP.get(
                frozenset(featureset) if featureset else featureset,
                featureset,
            )
        if featureset is None:
            featureset = self.NONEFEATURE
        ret[self.features.index(featureset)] = 1
        return ret

    def set2cls(self, suba_location):
        ret = [0] * len(self.classes_)
        if not suba_location:
            return ret  # if we a picking up *all* data
        for s in suba_location:
            s = self.Cat_[s]
            try:
                ret[self.classes_.index(s)] = 1
            except ValueError:
                print("no category", s, file=sys.stderr)
                raise
        return ret

    def gen_data(self, gold_std):
        p = self.predictor_
        # SIC outerjoin!
        j = gold_std.outerjoin(p, gold_std.c.locus == p.c.locus)
        col = (
            gold_std.c.suba_location
            if "suba_location" in gold_std.c
            else literal_column("0", type_=Integer).label("gs_location")
        )
        q = select([gold_std.c.locus, col, p.c.suba_location])
        q = q.select_from(j)

        if self.ignore_none_:
            q = q.where(p.c.suba_location != null())  # outer join may pick up None

        if VERBOSE:
            print(self.name, sql2str(q), file=sys.stderr)
        for locus, gsloc, ploc in tair10_engine.execute(q):
            x = self.set2vec(ploc)
            y = self.set2cls(gsloc)
            yield locus, x, y

    def consensus(self, tair10):
        for locus, x, y in self.gen_data(tair10):
            assert y == 0 or sum(y) == 0, (locus, y)  # tair10 has no suba_location
            yield locus, self.predict_v(x)

    def fit_data(self, data, alpha=1.0):
        NF = len(self.features)
        NC = len(self.classes_)  # number of columns (categories/locations) for Y
        ND = len(data)  # number of rows
        X = np.zeros((ND, NF), dtype=np.double)
        Y = np.zeros((ND, NC), dtype=np.double)

        check = set()
        # data order doesn't matter
        for irow, (locus, xrow, yrow) in enumerate(data):
            lu = locus.upper()
            assert lu not in check, locus
            check.add(lu)
            Y[irow, :] = yrow
            X[irow, :] = xrow

        self.YTX = np.dot(Y.T, X)
        self.X = X
        self.Y = Y
        assert self.YTX.shape == (len(self.classes_), NF)

        self._update_alpha(alpha)

    def effective_feature_count(self):
        return len(self.features)

    def _update_alpha(self, alpha):
        self.alpha_ = alpha
        M = np.sum(self.YTX, axis=1)  # sum row (sum over features)
        M = M[:, np.newaxis]  # (ncats,1)
        # calculate the laplacian correction
        self.M_ = M + self.effective_feature_count() * alpha
        self.fk_ = (self.YTX + alpha) / self.M_
        self.logfk_ = np.log(self.fk_)
        self.logfk_m1_ = np.log(1.0 - self.fk_)

    def init_fit(self, tair10="tair10"):
        self.get_features_(Meta3.tables[tair10] if isinstance(tair10, str) else tair10)

    def fit(self, gold_std, alpha=1.0):
        data = list(self.gen_data(gold_std))
        self.fit_data(data, alpha)
        return data

    def predict_v(self, features):
        # vs is like a selector since it is *usually* of
        # the form (0,0,1,..,0)

        vf = np.dot(self.logfk_, features)
        return vf

    def predict_from_agi(self, agi):
        def predict_(featureset):
            if not featureset and self.ignore_none_:
                return 0.0  # log of 1 is zero
            vs = self.set2vec(featureset)
            return self.predict_v(vs)

        p = self.predictor_
        q = select([p.c.suba_location]).where(p.c.locus == agi)
        r = tair10_engine.execute(q).fetchall()
        if not r:
            return predict_(None)
        assert len(r) == 1, r
        return predict_(r[0].suba_location)

    # some methods to display the data

    def rst(self, fmt=str, prefix="    ", SEP="\t"):
        def ffmt(f):
            if isinstance(f, (set, frozenset)):
                return ",".join(fmt(x) for x in sorted(f))
            return fmt(f)

        header = ["categories"] + [ffmt(f) for f in self.features]
        table = []
        for cat, row in zip(self.classes_, self.YTX):
            table.append([cat] + [fmt(f) for f in row])

        rsttable(table, header=header, SEP=SEP, prefix=prefix)

    def rstfk(self, fmt=lambda f: "%3f" % f, prefix="    ", SEP="\t"):
        def ffmt(f):
            if isinstance(f, (set, frozenset)):
                return ",".join(fmt(x) for x in sorted(f))
            return fmt(f)

        h = ["categories"] + [ffmt(f) for f in self.features]
        b = []
        for cat, row in zip(self.classes_, self.fk_):
            row = [cat] + [fmt(f) for f in row]
            b.append(row)

        rsttable(b, header=h, prefix=prefix, SEP=SEP)

    def showpredictor(self, gs, alpha=1.0, prefix="    ", SEP="\t", fmt=str):
        def ffmt(f):
            if isinstance(f, (set, frozenset)):
                return ",".join(fmt(x) for x in sorted(f))
            return fmt(f)

        self.init_fit("tair10")
        self.fit(gs, alpha=alpha)

        table = []
        header = ["categories"] + [ffmt(f) for f in self.features] + ["sum"]
        csum = np.zeros(len(self.features))
        for cat, row in zip(self.classes_, self.YTX):
            csum += row
            table.append([cat] + [int(r) for r in row] + [int(sum(row))])
        table.append(["sum"] + [int(r) for r in csum] + [int(sum(csum))])
        rsttable(table, header=header, SEP=SEP, prefix=prefix)

        table = []
        csum = np.zeros(len(self.features))
        for cat, row in zip(self.classes_, self.fk_):
            csum += row
            table.append([cat] + ["%.3f" % (r) for r in row] + ["%.3f" % sum(row)])
        table.append(["sum"] + ["%.3f" % (r) for r in csum] + ["%.3f" % (sum(csum))])
        rsttable(table, header=header, SEP=SEP, prefix=prefix)

    def remap(self):
        """create remappings for predictor featureset and gold standard feature set self.classes_"""
        K = list(self.features[:])
        imap = {}
        # map gs classes to predictor classes
        # we don't need to remap the features since
        # we keep the same order except for adding None to the end
        fn = self.NONEFEATURE
        for idx, c in enumerate(self.classes_):
            hit = False
            for jdx, kset in enumerate(K):
                if c in kset:
                    imap[idx] = jdx
                    hit = True
                    break
                elif c == "secretory" and len(kset & SECRECTORY) > 0:
                    imap[idx] = jdx
                    hit = True
                    break
            if hit:
                continue

            fc = frozenset([c])

            if fc in K:
                imap[idx] = K.index(fc)
            elif not self.ignore_none_:
                if fn in K:  # K has an unknown category (we map to that)
                    imap[idx] = K.index(fn)
                else:
                    K.append(fn)  # add unknown category
                    imap[idx] = K.index(fn)  # and map that

        NK = len(K)

        def mapperGS(yrow):
            ret = np.zeros(NK)
            for idx, v in enumerate(yrow):
                ret[imap[idx]] += v
            return ret

        def mapperP(xrow):
            ret = np.zeros(NK)
            for i, v in enumerate(xrow):
                ret[i] = v
            return ret

        return imap, K, mapperP, mapperGS

    def mi_mcc(self):
        def feat2str(f):
            if isinstance(f, str):
                return f
            if f is None:
                return str(f)
            return ",".join(map(str, sorted(f)))  # probably a set

        imap, K, mapP, mapGS = self.remap()

        NC = len(K)
        fn = self.NONEFEATURE
        NoneIdx = K.index(fn) if fn in K else -1
        conf = np.zeros((NC, NC))
        for idx in range(len(self.Y)):
            features = self.X[idx]
            yrow = self.Y[idx]
            yrow = mapGS(yrow)
            probs = mapP(features)
            assert len(probs) == len(yrow) and len(probs) == NC, (probs, yrow)

            if (
                self.ignore_none_ and NoneIdx >= 0 and np.sum(probs) == probs[NoneIdx]
            ):  # mapped only to None
                continue
            conf += np.outer(yrow, probs)

        fmt = "%.2f"
        # fmt = '%s'

        fcols = np.sum(conf, axis=0)
        frows = np.sum(conf, axis=1)

        confx = [
            [feat2str(cat)] + [fmt % r for r in row] + [fmt % frows[idx]]
            for idx, (cat, row) in enumerate(zip(K, conf))
        ]
        confx.append(["SUM"] + [fmt % r for r in fcols] + [fmt % np.sum(conf)])

        print()
        print("predictor:\t", self.name)
        print()
        print("mapped categories:\t", "\t".join(feat2str(s) for s in K))
        print("features:\t", "\t".join(feat2str(s) for s in self.features))
        print("categories:\t", "\t".join(feat2str(s) for s in self.classes_))
        print("cat 2 feature mapping:\t", imap)
        print()
        rsttable(
            confx,
            header=["location"] + [feat2str(s) for s in K] + ["SUM"],
            SEP="\t",
            prefix="\t",
        )

        mi = Phase0.MI(conf)
        mcc = dict(Phase0.MCC(conf, K))

        # cc=list(self.classes_)+[None]
        H = ["location"] + list(Phase1.STAT)
        print()
        print("MI\t", mi)
        print()
        rows = []
        for location in K:
            row = [feat2str(location)] + [
                ("%.3f" % mcc[location][s] if location in mcc else "")
                for s in Phase1.STAT
            ]
            rows.append(row)

        rsttable(rows, header=H, SEP="\t", prefix="\t")


class MultiPredictor(Predictor):
    FMAP = None
    FMAP1 = {
        "golgi": "secretory",
        "plasma membrane": "secretory",
        "vacuole": "secretory",
        "endoplasmic reticulum": "secretory",
        "extracellular": "secretory",
        "cytoskeleton": "cytosol",
        "unclear": None,
    }
    FMAP2 = {
        "cytoskeleton": None,
        "unclear": None,
        "mitochondrion": None,
        "plastid": None,
        "nucleus": None,
        "cytosol": None,
        "peroxisome": None,
    }
    FMAPALL = {"cytoskeleton": "cytosol"}

    NONEFEATURE = None

    def feat2str(self):
        return [str(f) for f in self.features]

    def get_features_(self, tair10):
        p = self.predictor_
        q = select([p.c.suba_location.distinct()])
        # we'll pick up nulls for rows that predictor doesn't have
        q = q.select_from(tair10.outerjoin(p, tair10.c.locus == p.c.locus))
        if self.ignore_none_:
            q = q.where(p.c.suba_location != null())

        features = set()
        for (locs,) in tair10_engine.execute(q):
            if locs is None:
                locs = [None]
            for s in locs:
                x = self.FMAP.get(s, s) if self.FMAP else s
                if x is None and self.ignore_none_:
                    continue
                features.add(x)

        features = sorted(features)

        # put None at end
        if None in features:
            i = features.index(None)
            del features[i]
            features = features + [None]

        self.features = tuple(features)
        print(
            self.__class__.__name__,
            "predictor:",
            self.name,
            "ignore none:",
            self.ignore_none_,
            "features:",
            self.features,
            file=sys.stderr,
        )

    def effective_feature_count(self):
        return 2  #

    def set2vec(self, suba_location):
        ret = [0] * len(self.features)
        if suba_location is None:
            suba_location = [None]

        for f in suba_location:
            try:
                ret[self.features.index(self.FMAP.get(f, f) if self.FMAP else f)] = 1
            except ValueError as e:
                raise ValueError(
                    f"{self.__class__.__name__}: no {f} in {self.features}",
                ) from e
        return ret

    def predict_v(self, features):
        vs = features

        vn = np.array([(0 if v else 1) for v in vs], dtype=np.double)
        try:
            return np.dot(self.logfk_, vs) + np.dot(self.logfk_m1_, vn)
        except FloatingPointError as e:
            print(sys.stderr, "RuntimeWaring", e, file=sys.stderr)
            self.rstfk(fmt=lambda f: "%.6f" % f)
            print(vs)
            print(vn)
            raise

    def remap(self):
        """create remappings for predictor featureset and gold standard feature set self.classes_"""
        K = list(self.features[:])

        imap = {}
        # map gs classes to predictor classes
        # we don't need to remap the features since
        # we keep the same order except for adding None to the end
        fn = None
        for idx, c in enumerate(self.classes_):
            fc = c
            if fc in K:
                imap[idx] = K.index(fc)
            elif not self.ignore_none_:
                if fn in K:  # K has an unknown category (we map to that)
                    imap[idx] = K.index(fn)
                else:
                    K.append(fn)  # add unknown category
                    imap[idx] = K.index(fn)  # and map that

        NK = len(K)

        def mapperGS(yrow):
            ret = np.zeros(NK)
            for idx, v in enumerate(yrow):
                ret[imap[idx]] += v
            return ret

        def mapperP(xrow):
            ret = np.zeros(NK)
            for i, v in enumerate(xrow):
                ret[i] = v
            return ret

        return imap, K, mapperP, mapperGS


class NumPredictor(MultiPredictor):
    FMAP = None
    _query = None
    # Phase1
    FMAP1 = {
        "golgi": "secretory",
        "endoplasmic reticulum": "secretory",
        "extracellular": "secretory",
        "vacuole": "secretory",
        "plasma membrane": "secretory",
        "cytosol": "cytosol",
        "cytoskeleton": "cytosol",
    }
    # Phase2
    FMAP2 = {
        "cytoskeleton": None,
        "unclear": None,
        "mitochondrion": None,
        "plastid": None,
        "nucleus": None,
        "cytosol": None,
        "peroxisome": None,
    }
    FMAPALL = {"cytoskeleton": "cytosol"}

    def get_features_(self, tair10):
        p = self.predictor_
        q = select([p.c.location.distinct()])
        # we'll pick up nulls for rows that predictor doesn't have
        q = q.select_from(tair10.outerjoin(p, tair10.c.locus == p.c.locus))
        if self.ignore_none_:
            q = q.where(p.c.location != null())
        q = q.where(p.c.location != "unclear")

        features = set()
        for (loc,) in tair10_engine.execute(q):
            assert loc is not None
            if self.FMAP:
                loc = self.FMAP.get(loc, loc)
            if loc is None and self.ignore_none_:
                continue
            features.add(loc)

        self.features = list(sorted(features))
        # place None at end
        if None in self.features:
            self.features.remove(None)
            self.features += [None]

        self.features = tuple(self.features)
        print(
            self.__class__.__name__,
            "predictor:",
            self.name,
            "ignore none:",
            self.ignore_none_,
            "features:",
            self.features,
            file=sys.stderr,
        )

    @property
    def query_(self):
        if hasattr(self, "_query"):
            return self._query
        # tables such as src_gfp are a list of
        # unique locus,location,PMID
        # we turn it into a "virtual" table of
        # locus,num papers in mito,num papers in plastid,...]
        # i.e. with a *unique* locus per row
        #
        # First map "feature" to list of locations
        imap = defaultdict(set)
        for loc, feature in self.FMAP.items():
            imap[feature].add(loc)
        # add the rest of the features
        for f in self.features:
            if f not in imap:
                imap[f].add(str(f))

        p = self.predictor_
        # this is in the order of features
        fc = [
            func.sum(
                case(value=p.c.location, whens={loc: 1 for loc in imap[f]}, else_=0),
            ).label(str(f))
            for f in self.features
        ]

        self._query = q = (
            select([p.c.locus] + fc)
            .where(p.c.location != "unclear")
            .group_by(p.c.locus)
            .alias("query")
        )
        return q

    def coverage(self):
        q = self.predictor_
        gs = self.gold_std_
        j1 = gs.join(q, gs.c.locus == q.c.locus)
        j2 = q.outerjoin(gs, gs.c.locus == q.c.locus)
        s = select([func.count(q.c.locus.distinct())])
        (ngs,) = s.select_from(j1).execute().fetchone()
        (nex,) = s.select_from(j2).where(gs.c.locus == null()).execute().fetchone()
        return ngs, nex

    def gen_data(self, gold_std):
        q = self.query_
        j = gold_std.join(q, gold_std.c.locus == q.c.locus)
        # add a "fake" suba_location column
        col = (
            gold_std.c.suba_location
            if "suba_location" in gold_std.c
            else literal_column("0", type_=Integer).label("suba_location")
        )
        q1 = select([gold_std.c.locus, col] + [q.c[str(f)] for f in self.features])
        q1 = q1.select_from(j)
        if VERBOSE:
            print(self.name, sql2str(q1), file=sys.stderr)

        for row in tair10_engine.execute(q1):
            gsloc = row[1]  # might be 0
            ploc = row[2:]
            x = list(ploc)  # this is all we need for set2vec
            y = self.set2cls(gsloc)
            yield row.locus, x, y

    def effective_feature_count(self):
        return len(self.features)

    def predict_v(self, features):
        M = sum(features)
        vf = np.dot(self.logfk_, features)
        return vf + M * np.log(self.M_) - self.M_

    def predict_from_agi(self, agi):
        q = self.query_
        q = q.where(q.c.locus == agi)
        r = tair10_engine.execute(q).fetchall()
        if not r:
            return 0
        assert len(r) == 1, r
        return self.predict_v(r[0][1:])


class PPIPredictor(NumPredictor):
    def __init__(self, *args, **kw):
        NumPredictor.__init__(self, *args, **kw)
        self._mkpredictor()

    def _mkpredictor(self):
        p = self.predictor_
        gs = self.gold_std_
        j = gs.join(p, p.c.locusA == gs.c.locus)
        q = select(
            [
                p.c.locusB.label("locus"),
                gs.c.suba_location,
                func.count(p.c.paper.distinct()).label("num"),
            ],
        )
        q = q.select_from(j)

        q = q.where(p.c.locusA != p.c.locusB).group_by(p.c.locusB, gs.c.suba_location)
        # save original table
        self._orig = self.predictor_
        # table of (locus,suba_location),count

        self.predictor_ = q.alias("pred_src_ppi")  # make pred_src_ppi the "predictor"

    @property
    def query_(self):
        if hasattr(self, "_query"):
            return self._query
        imap = defaultdict(set)
        if self.FMAP:
            for loc, feature in self.FMAP.items():
                imap[feature].add(loc)
        for f in self.features:
            if f not in imap:
                imap[f].add(str(f))
        p = self.predictor_
        fc = [
            func.sum(
                func.if_(
                    or_(
                        *(
                            func.find_in_set(loc, p.c.suba_location) > 0
                            for loc in imap[f]
                        ),
                    ),
                    p.c.num,
                    0,
                ),
            ).label(str(f))
            for f in self.features
        ]

        q = select([p.c.locus] + fc).group_by(p.c.locus)
        self._query = q = q.alias("query")
        return q

    def get_features_(self, tair10):
        p = self.predictor_
        features = set()
        q = select([p.c.suba_location.distinct()]).where(p.c.suba_location != null())
        for (suba_location,) in tair10_engine.execute(q):
            for s in suba_location:
                s = self.FMAP.get(s, s) if self.FMAP else s  # map secretory
                features.add(s)
        self.features = tuple(sorted(features))
        print(
            self.__class__.__name__,
            "predictor:",
            self.name,
            "ignore none:",
            self.ignore_none_,
            "features:",
            self.features,
            file=sys.stderr,
        )


class AttedPredictor(PPIPredictor):
    def __init__(self, *args, **kw):
        rank = kw.pop("rank")
        avg = kw.pop("avg")
        self.rank_ = rank  # these are category to rank mapping
        self.avg_ = avg
        PPIPredictor.__init__(self, *args, **kw)

    def _mkpredictor(self):
        self._orig = self.predictor_
        q = self.filter_(self.predictor_, self.rank_, self.avg_)
        # will have multiple locus with different suba_locations
        self.predictor_ = q.alias("query")  # the "real" predictor table

    def filter_(self, p, rank, avg):
        mygs = self.gold_std_
        p = self._orig
        imap = defaultdict(set)
        for c in self.Cat_:
            cls = self.Cat_[c]
            imap[cls].add(c)

        filt = or_(
            *(
                and_(
                    or_(
                        *(
                            func.find_in_set(loc, mygs.c.suba_location) > 0
                            for loc in imap[cat]
                        ),
                    ),  # location is correct for this category
                    and_(
                        p.c.rank < rank[cat],
                        p.c.ave > avg[cat],
                    ),  # and rank is correct
                )
                for cat in imap
            ),
        )
        j = mygs.join(p, p.c.locus == mygs.c.locus)
        q = select(
            [
                p.c.locus_coex.label("locus"),
                mygs.c.suba_location,
                func.count(text("distinct rank, ave")).label("num"),
            ],
        )
        q = q.select_from(j)
        q = q.where(p.c.rank >= 0)
        q = q.where(filt)
        q = q.where(p.c.locus != p.c.locus_coex).group_by(
            p.c.locus_coex,
            mygs.c.suba_location,
        )
        return q

    def coverage(self):
        q = self.filter_(self._orig, self.rank_, self.avg_).alias("q")
        gs = self.gold_std_
        j1 = gs.join(q, gs.c.locus == q.c.locus)
        j2 = q.outerjoin(gs, gs.c.locus == q.c.locus)
        s = select([func.count(q.c.locus.distinct())])
        (ngs,) = s.select_from(j1).execute().fetchone()
        (nex,) = s.select_from(j2).where(gs.c.locus == null()).execute().fetchone()
        return ngs, nex


Avg = {
    "secretory": 0.5,
    "peroxisome": 0.6,
    "nucleus": 0.5,
    "mitochondrion": 0.6,
    "cytosol": 0.8,
    "plastid": 0.8,
    "vacuole": 0.7,
    "endoplasmic reticulum": 0.8,
    "golgi": 0.7,
    "extracellular": 0.7,
    "cytoskeleton": 0.7,
    "plasma membrane": 0.7,
}
# the default predictor
DefPred = lambda C, p, gs: Predictor(Meta3.tables[p], C, ignore_none=False, gold_std=gs)
# variations....
PMAP = {
    "pred_plantmploc": lambda C, p, gs: MultiPredictor(
        Meta3.tables[p],
        C,
        ignore_none=False,
        gold_std=gs,
    ),
    "src_gfp": lambda C, p, gs: NumPredictor(
        Meta3.tables[p],
        C,
        ignore_none=True,
        gold_std=gs,
    ),
    "src_msms": lambda C, p, gs: NumPredictor(
        Meta3.tables[p],
        C,
        ignore_none=True,
        gold_std=gs,
    ),
    "src_ppi": lambda C, p, gs: PPIPredictor(
        Meta3.tables[p],
        C,
        ignore_none=True,
        gold_std=gs,
    ),
    "atted_0": lambda C, p, gs: AttedPredictor(
        Meta3.tables[p],
        C,
        ignore_none=True,
        gold_std=gs,
        rank=defaultdict(lambda: 50),
        avg=Avg,
    ),
}


# a container class for a list of predictors
class PredictorsAll:
    def __init__(self, gold_std, Phase, use_exp=True, use_pred=True):
        C = Phase.getCategories(gold_std)
        self.classes_ = tuple(sorted(set(C.values())))
        self.Cat_ = C
        preds = []
        if use_pred:
            if use_pred == "all":
                _preds = Phase.getPredictors()
            else:
                _preds = [
                    Meta3.tables[n]
                    for n in [
                        (m if m.startswith("pred_") else "pred_" + m)
                        for m in use_pred.split(",")
                    ]
                ]
            for p in _preds:
                p = PMAP.get(p.name, DefPred)(C, p.name, gold_std)
                preds.append(p)
        if use_exp:
            if use_exp == "all":
                preds.extend(
                    [
                        PMAP[n](C, n, gold_std)
                        for n in ("src_msms", "src_gfp", "src_ppi", "atted_0")
                    ],
                )
            else:
                for n in use_exp.split(","):
                    preds.append(PMAP[n](C, n, gold_std))

        self.predictors_ = preds
        self.gold_std_ = gold_std
        print(self.__class__.__name__, "Predictors", file=sys.stderr)
        for p in self.predictors_:
            print(Phase.__name__, p.name, file=sys.stderr)

        # normalize the priors
        _s = sum(Phase.PRIORS[k] for k in self.classes_)
        self.log_priors_ = np.log(
            np.array([Phase.PRIORS[k] / _s for k in self.classes_]),
        )

        # which phase feature mapping
        for p in self.predictors_:
            p.FMAP = p.FMAP1

    def mi_mcc(self):
        for p in self.predictors_:
            p.mi_mcc()

    def set2cls(self, suba_location):
        ret = [0] * len(self.classes_)
        if not suba_location:
            return ret  # if we a picking up *all* data
        for s in suba_location:
            s = self.Cat_[s]
            ret[self.classes_.index(s)] = 1
        return ret

    def init_fit(self, tair10="tair10"):
        for p in self.predictors_:
            p.init_fit(tair10)

    def fit(self, gold_std, alpha=1.0):
        for p in self.predictors_:
            p.fit(gold_std, alpha)

    def init_leave1out(self, tair10, gold_std):
        for p in self.predictors_:
            p.init_leave1out(tair10, gold_std)

    def leave1out(self, locus, alpha):
        ret = []
        for p in self.predictors_:
            try:
                probs = p.leave1out(locus, alpha)
                ret.append(probs)
            except ValueError:
                print("skipping", locus, "for", p.name, file=sys.stderr)

        probs = sum(ret)
        p = np.exp(probs + self.log_priors_)
        probs = p / np.sum(p)
        # assert fabs(sum(probs)-1)<1.0e-6,probs
        return dict(zip(self.classes_, probs))

    def calls(self, probs, fract=0.5):
        s = 0.0
        z = np.zeros(len(self.classes_))
        for c, p in sorted(zip(self.classes_, probs), reverse=True, key=lambda x: x[1]):
            z[self.classes_.index(c)] = 1
            s += p
            if s > fract:
                return z
        return z

    def consensus(self, tair10):
        d = defaultdict(list)
        for p in self.predictors_:
            print("consensus:", p.name, file=sys.stderr)
            for locus, probs in p.consensus(tair10):
                d[locus].append(probs)

        for locus in sorted(d):
            npred = len(d[locus])  # number of predictors that participated
            probs = sum(d[locus])  # sic not np.sum!
            p = np.exp(probs + self.log_priors_)
            p = p / np.sum(p)
            yield locus, p, npred

    def load_consensus(
        self,
        tablename,
        gold_std="gold_std",
        tair10="tair10",
        fract=0.5,
    ):
        tair10 = Meta3.tables[tair10]
        gold_std = Meta3.tables[gold_std]

        if tablename in Meta3.tables:
            target = Meta3.tables[tablename]
        else:
            target = mk_consensus_table(tablename, meta3=Meta3, locnames=self.classes_)

        if target.exists(bind=tair10_engine):
            print("dropping", target.name, file=sys.stderr)
            target.drop(bind=tair10_engine)

        print("creating", target.name, file=sys.stderr)
        target.create(bind=tair10_engine)

        data = []

        def insert(n=5000):
            if len(data) < n:
                return 0
            proxy = tair10_engine.execute(target.insert(), *data)
            assert proxy.rowcount == len(data)
            data[:] = []
            print(nn + proxy.rowcount, "done", file=sys.stderr)
            return proxy.rowcount

        nn = 0
        print("loading....", tablename, file=sys.stderr)

        for locus, probs, _ in self.consensus(tair10):
            store = {n: float(p) for n, p in zip(self.classes_, probs)}
            vals = list(sorted(store.iteritems(), key=lambda x: x[1], reverse=True))
            store["locus"] = locus
            # store['num']=1
            for i, n in enumerate(self.classes_):
                store["call%d" % (i + 1)] = vals[i][0]
                store["prob%d" % (i + 1)] = vals[i][1]

            sl = set()
            tot = 0.0
            for n, p in vals:
                n = Phase0.S2ENUM.get(n, n)
                sl.add(n)
                tot += p
                if tot > fract:
                    break

            assert len(sl) >= 1, sl
            store["suba_location"] = sl
            data.append(store)
            nn += insert()

        nn += insert(1)
        print("ok:", nn, "entries", file=sys.stderr)
        return nn


class PredictorsAll2(PredictorsAll):
    def __init__(self, *args, **kw):
        PredictorsAll.__init__(self, *args, **kw)
        for p in self.predictors_:
            p.FMAP = p.FMAP2


class PredictorsAllC(PredictorsAll):
    def __init__(self, *args, **kw):
        PredictorsAll.__init__(self, *args, **kw)
        for p in self.predictors_:
            p.FMAP = p.FMAPALL


class Phase0:
    CYTOSOL = ("cytosol", "cytoskeleton")
    SECRETORY = (
        "golgi",
        "endoplasmic reticulum",
        "extracellular",
        "vacuole",
        "plasma membrane",
    )
    S2ENUM = {
        "secretory": "golgi",
    }  # we map secretory back to golgi so we can store it in a suba_location column
    INCLUDE = None  # means include all
    PRIORS = {}

    @classmethod
    def getPredictors(cls):
        ret = []
        for t in Meta3.tables.itervalues():
            if not t.exists():
                continue
            if not t.name.startswith("pred_"):
                continue
            if "suba_location" not in t.c:
                continue  # not a predictor table
            # pylint: disable=unsupported-membership-test
            if cls.INCLUDE is not None and t.name not in cls.INCLUDE:
                continue
            ret.append(t)

        return sorted(ret, key=lambda t: t.name)

    # map input locations to Bayes locations
    @classmethod
    def _mapcats(cls, locs):
        suba_location = locs
        assert suba_location  # can't be null
        for loc in suba_location:
            if loc == "unclear":
                continue
            elif loc in cls.SECRETORY:
                yield (loc, "secretory")
            elif loc in cls.CYTOSOL:
                yield (loc, "cytosol")
            else:
                if loc in cls.PRIORS:
                    yield (loc, loc)
                else:
                    yield (loc, None)  # make say cytoskeleton to "unknown"

    @classmethod
    def getCategories(cls, gold_std):
        def _mkcategories(suba_location):
            # subloc might filter out all predictions e.g. 'unclear' and hence be empty
            cats = dict(cls._mapcats(suba_location)) if suba_location else None
            cats = cats if cats else {None: None}  # we always produce a non-empty set
            return cats

        def _showmap(gold_std, D):
            V = sorted(set(D.values()))
            print(
                cls.__name__,
                gold_std.name,
                ",".join(str(v) for v in V),
                file=sys.stderr,
            )

        q = select([gold_std.c.suba_location.distinct()])
        K = {}
        for r in tair10_engine.execute(q):
            cd = _mkcategories(r.suba_location)
            K.update(cd)
        _showmap(gold_std, K)
        return K

    @staticmethod
    def MCC(conf, classes):
        assert conf.shape == (len(classes), len(classes)), (conf.shape, classes)
        TOTAL = np.sum(conf)
        conf = conf / TOTAL
        for i, cat in enumerate(classes):
            TP = conf[i, i]  #
            FP = np.sum(conf[:, i]) - TP
            FN = np.sum(conf[i, :]) - TP
            TN = 1.0 - TP - FP - FN
            DENOM = (TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)
            MCC = (TP * TN - FP * FN) / sqrt(float(DENOM)) if DENOM else 0.0
            PREC = TP / float(TP + FP) if TP + FP else 0.0  # positive predictive value
            SPEC = (
                TN / float(TN + FP) if TN + FP else 0.0
            )  # specificity # prob of negative test give patient is well
            RECALL = TP / float(TP + FN) if TP + FN else 0.0  # sensitivity
            NPV = TN / float(FN + TN) if FN + TN else 0.0  # negative predictive value
            F1 = 2.0 * PREC * RECALL / (PREC + RECALL) if PREC + RECALL else 0.0
            yield cat, dict(
                MCC=MCC,
                PREC=PREC,
                RECALL=RECALL,
                SPEC=SPEC,
                TOTAL=TOTAL,
                NPV=NPV,
                TP=TP,
                TN=TN,
                FP=FP,
                FN=FN,
                F1=F1,
            )

    @staticmethod
    def MI(matrix):
        T = np.sum(matrix)
        z = np.zeros(matrix.shape)
        assert np.all(np.greater_equal(matrix, z)), matrix
        assert T > 0, (T, matrix)
        conf = matrix / T
        assert np.all(np.greater_equal(conf, z)), conf
        row2 = np.array(
            [np.sum(conf[i, :]) for i in range(conf.shape[0])],
            dtype=np.double,
        )
        col2 = np.array(
            [np.sum(conf[:, i]) for i in range(conf.shape[1])],
            dtype=np.double,
        )
        row = np.sum(conf, axis=1)  # ,dtype=np.double)
        assert np.allclose(row, row2), (row, row2)
        col = np.sum(conf, axis=0)  # ,dtype=np.double)
        assert np.allclose(col, col2), (col, col2)
        fifj = np.outer(row2, col2)  # sum[:,i]*sum[j,:]
        assert np.all(np.greater_equal(fifj, z)), (fifj, conf, row, col)
        mi = 0.0
        for i in range(conf.shape[0]):
            for j in range(conf.shape[1]):
                f = conf[i, j]
                if f == 0.0:
                    continue  # 0log0 is 0
                assert f > 0 and fifj[i, j] > 0, (
                    i,
                    j,
                    f,
                    fifj[i, j],
                    fifj,
                    conf,
                    row,
                    col,
                )
                mi += f * log(f / fifj[i, j])
        return mi


class Phase1(Phase0):
    PRIORS = {
        "mitochondrion": 7.3,
        "plastid": 14.6,
        "cytosol": 18,
        "peroxisome": 1.6,
        "nucleus": 35,
        "secretory": 1.7 + 3.2 + 11.8 + 19 + 3.6,
    }

    STAT = ("MCC", "SPEC", "PREC", "RECALL", "NPV", "TP", "TN", "FP", "FN", "TOTAL")

    NEED_NONE = (
        "pred_atp",
        "pred_chlorop",
        "pred_mitopred",
        "pred_mitoprot2",
        "pred_pts1",
        "pred_nucleo",
        "pred_pclr",
        "pred_ipsort",  # all null values are like not secretory/mito/plastid
    )


class Phase2(Phase1):
    INCLUDE = (
        "pred_adaboost",
        "pred_epiloc",
        "pred_multiloc2",
        "pred_plantmploc",
        "pred_wolfpsort",
        "pred_yloc",
        "src_msms",
        "src_gfp",
    )

    PRIORS = {
        "golgi": 1.0,
        "endoplasmic reticulum": 3.0,
        "extracellular": 10.0,
        "vacuole": 3.0,
        "plasma membrane": 16.0,
    }

    @classmethod
    def _mapcats(cls, locs):
        for loc in locs:
            if loc == "unclear":
                continue
            elif loc in cls.CYTOSOL:
                yield (loc, "cytosol")
            else:
                if loc in cls.PRIORS:
                    yield (loc, loc)
                else:
                    yield (loc, None)  # make say cytoskeleton to "unknown"


class Combined(Phase1):
    INCLUDE = None  # this means include all predictors!

    PRIORS = {
        "cytosol": 0.15544041450777202,
        "endoplasmic reticulum": 0.030852567121997177,
        "extracellular": 0.10284189040665725,
        "golgi": 0.010284189040665725,
        "mitochondrion": 0.06303972366148532,
        "nucleus": 0.3022452504317789,
        "peroxisome": 0.01381692573402418,
        "plasma membrane": 0.1645470246506516,
        "plastid": 0.12607944732297063,
        "vacuole": 0.030852567121997177,
    }

    @classmethod
    def _mapcats(cls, locs):
        for loc in locs:
            if loc == "unclear":
                continue
            elif loc in cls.CYTOSOL:
                yield (loc, "cytosol")
            else:
                if loc in cls.PRIORS:
                    yield (loc, loc)
                else:
                    yield (loc, None)  # map say cytoskeleton to "unknown"

    @classmethod
    def combine(
        cls,
        tablename="subacon",
        phase1="subacon_p1",
        phase2="subacon_p2",
        percent=0.5,
        merge="secretory",
    ):
        phase1 = Meta3.tables[phase1]
        phase2 = Meta3.tables[phase2]

        p1 = list(tair10_engine.execute(phase1.select()))
        p2 = {r.locus.upper(): r for r in tair10_engine.execute(phase2.select())}
        # check a column is a location/category column

        def iscat(c):
            if c.name in ("locus", "suba_location", "num"):
                return False
            if c.name.startswith(("call", "prob")):
                return False
            return True

        phase1Cats = tuple(sorted(c.name for c in phase1.c if iscat(c)))
        phase2Cats = tuple(sorted(c.name for c in phase2.c if iscat(c)))

        assert merge in phase1Cats, (merge, phase1Cats)
        assert merge not in phase2Cats, (merge, phase2Cats)

        # create the final location hame
        Locnames = set(phase1Cats + phase2Cats)
        Locnames.remove(merge)
        Locnames = tuple(sorted(Locnames))
        # sanity check
        assert set(Locnames) == set(cls.PRIORS), Locnames

        p_combined = []

        # merge the results
        for r in p1:
            probs1 = {n: r[n] for n in phase1Cats}
            s = sum(probs1.values())
            assert fabs(s - 1.0) < 1.0e-5, s  # all sum to 1 because probability
            r2 = p2[r.locus.upper()]
            probs2 = {n: r2[n] for n in phase2Cats}
            s = sum(probs2.values())
            assert fabs(s - 1.0) < 1.0e-5, s  # all sum to 1 because probability

            pmerge = probs1.pop(merge)

            for n in probs2:
                assert n not in probs1, n
                probs1[n] = probs2[n] * pmerge

            s = sum(probs1.values())
            assert fabs(s - 1.0) < 1.0e-5, s  # all sum to 1
            # now order picks
            vals = list(sorted(probs1.items(), reverse=True, key=lambda kv: kv[1]))

            for idx, (loc, prob) in enumerate(vals):
                probs1["call%d" % (idx + 1)] = loc
                probs1["prob%d" % (idx + 1)] = prob
            # create consensus
            locs = set()
            sval = 0.0
            for loc, p in vals:
                sval += p
                locs.add(loc)
                if sval > percent:
                    break
            probs1["suba_location"] = locs
            probs1["locus"] = r.locus
            p_combined.append(probs1)

        # create subacon table
        if tablename in Meta3.tables:
            ct = Meta3.tables[tablename]
            assert set(Locnames) == {c.name for c in ct.c if iscat(c)}
        else:
            ct = mk_consensus_table(tablename, meta3=Meta3, locnames=Locnames)
        if ct.exists(bind=tair10_engine):
            ct.drop(bind=tair10_engine)

        ct.create(bind=tair10_engine)

        # insert the data
        proxy = tair10_engine.execute(ct.insert(), *p_combined)

        print(proxy.rowcount, len(p_combined), file=sys.stderr)
        assert proxy.rowcount == len(p_combined)


def goldstd2sec(tablename="gold_std_sec_only", gold_std="gold_std"):
    # create the gold_std_sec_table
    if tablename in Meta3.tables:
        culled = Meta3.tables[tablename]
        culled.drop(bind=tair10_engine)
    tair10_engine.execute(
        f"create table if not exists `{tablename}` like `{gold_std}`;",
    )

    gs = Meta3.tables[gold_std]

    Meta3.reflect(only=[tablename])
    culled = Meta3.tables[tablename]
    mask = Mask(Phase2.PRIORS)  # create a mask for suba_location
    nmask = ~mask
    cols = [
        gs.c.locus,
        gs.c.suba_location.op("&")(mask).label("suba_location"),
        gs.c.notesCMH,
    ]
    q = select(cols).where(
        or_(*(func.find_in_set(s, gs.c.suba_location) > 0 for s in Phase2.PRIORS)),
    )

    p = tair10_engine.execute(culled.insert().from_select([c.name for c in cols], q))
    print("loaded", p.rowcount, "entries", file=sys.stderr)

    # find any that have extra locations not in Phase2.PRIORS and remove them
    # e.g. turn suba_location 'golgi,nucleus' into 'golgi'
    q = (
        select([culled])
        .where(culled.c.suba_location.op("&")(nmask) > 0)
        .where(
            or_(
                *(
                    func.find_in_set(s, culled.c.suba_location) > 0
                    for s in Phase2.PRIORS
                ),
            ),
        )
    )
    ids = []
    for row in tair10_engine.execute(q):
        print(row.locus, row.suba_location, file=sys.stderr)
        ids.append(row.id)
    if not ids:
        return
    u = (
        culled.update()
        .where(culled.c.id.in_(ids))
        .values({culled.c.suba_location: culled.c.suba_location.op("&")(mask)})
    )
    proxy = tair10_engine.execute(u)
    print(proxy.rowcount, len(ids), file=sys.stderr)
    # they are really gone
    assert len(tair10_engine.execute(q).fetchall()) == 0


def combine_phases(
    tablename="subacon",
    phase1="subacon_p1",
    phase2="subacon_p2",
    percent=0.5,
    merge="secretory",
):
    Combined.combine(
        tablename=tablename,
        phase1=phase1,
        phase2=phase2,
        percent=percent,
        merge=merge,
    )


def showpredictor(predictor, gold_std="gold_std", phase=1):
    gold_std = Meta3.tables[gold_std]
    Phase = Phase1 if phase == 1 else Phase2
    C = Phase.getCategories(gold_std)
    p = PMAP.get(predictor, DefPred)(C, predictor, gold_std)
    p.FMAP = p.FMAP1 if phase == 1 else p.FMAP2
    p.showpredictor(gold_std, alpha=1.0, prefix="\t", SEP="\t", fmt=str)
    p.mi_mcc()


def subacon(
    tablename="subacon",
    use_exp="",
    use_pred="",
    gold_std="gold_std",
    fract=0.5,
    phase="1",
):
    if use_pred:  # check
        if isinstance(use_pred, str):
            if use_pred != "all":
                use_pred = use_pred.split(",")
                for n in use_pred:
                    assert n in Meta3.tables, n
    else:
        use_pred = []
    if use_exp:  # check
        if isinstance(use_exp, str):
            if use_exp != "all":
                use_exp = use_exp.split(",")
                for n in use_exp:
                    assert n in Meta3.tables, n
    else:
        use_exp = []
    Phase = Phase1 if phase == "1" else Phase2
    PA = PredictorsAll if phase == "1" else PredictorsAll2
    gs = Meta3.tables[gold_std]
    p = PA(gs, Phase=Phase, use_exp=use_exp, use_pred=use_pred)
    p.init_fit(tair10="tair10")
    p.fit(gs)
    p.load_consensus(
        tablename=tablename,
        gold_std=gold_std,
        tair10="tair10",
        fract=fract,
    )


def dump():
    tables = list(PNAMES)
    tables = tables + ["tair10", "gold_std"]
    tables = " ".join(tables)
    host = tair10_engine.url.host
    user = tair10_engine.url.username
    port = tair10_engine.url.port or 3306
    password = tair10_engine.url.password

    print("dumping tables:", tables, file=sys.stderr)
    os.system(
        "time mysqldump --host=%s  --port=%d --user=%s --password=%s"
        " --max_allowed_packet=32M --single-transaction tair10 --tables %s | gzip > subacon.sql.gz"
        % (host, port, user, password, tables),
    )


# pylint: disable=redefined-outer-name
def merge_gold_std(subacon="subacon", gold_std="gold_std"):
    subacon = Meta3.tables[subacon]
    gold_std = Meta3.tables[gold_std]

    print([c.name for c in subacon.c])
    print(LOCNAMES)
    S = {c for c in LOCNAMES if c in subacon.c}

    print(S)

    def mkdict(loc):
        assert loc in S
        s = S.copy()
        s.remove(loc)
        locs = [loc] + list(sorted(s))
        V = {"suba_location": {loc}}
        n = 0
        for c in locs:
            v = 1.0 if c == loc else 0.0
            n += 1
            cn = "call%d" % (n)
            pn = "prob%d" % (n)
            V[pn] = v
            V[cn] = c
            V[c] = v
        return V

    V = mkdict("mitochondrion")
    u = subacon.update().where(subacon.c.locus.like("ATMG%")).values(V)
    p = tair10_engine.execute(u)
    print("mito", p.rowcount)
    V = mkdict("plastid")
    u = subacon.update().where(subacon.c.locus.like("ATCG%")).values(V)
    p = tair10_engine.execute(u)
    print("plastid", p.rowcount)

    ul = []
    for locus, locs in tair10_engine.execute(
        select([gold_std.c.locus, gold_std.c.suba_location]),
    ):
        s = S.copy()

        for l in locs:
            if l in s:
                s.remove(l)
            else:
                print("gs has", locus, l)
        V = {"suba_location": locs}
        V["agi"] = locus
        p = 1.0 / len(locs)
        n = 0
        for c in locs:
            n += 1
            cn = "call%d" % (n)
            pn = "prob%d" % (n)
            V[pn] = p
            V[cn] = c
            V[c] = p
        for c in sorted(s):
            n += 1
            cn = "call%d" % (n)
            pn = "prob%d" % (n)
            V[pn] = 0.0
            V[cn] = c
            V[c] = 0.0
        ul.append(V)

    u = subacon.update().where(subacon.c.locus == bindparam("agi"))
    p = tair10_engine.execute(u, *ul)
    print(p.rowcount)


def cmds(sub):
    # pylint: disable=no-member
    host = lambda aa: aa(
        "--host",
        default="mysql://127.0.0.1/suba3",
        help="default SUBAcon database",
    )

    f = dump
    ap = sub.add_parser(f.func_name, help="dump predictors for suba")
    ap.set_defaults(func=f)
    aa = ap.add_argument
    host(aa)

    f = showpredictor
    ap = sub.add_parser(f.func_name, help="show stats for predictor")
    ap.set_defaults(func=f)
    aa = ap.add_argument
    host(aa)
    aa("--predictor", help="predictor to compute", required=True)
    aa("--gold_std", help="gold standard table [%(default)s]", default="gold_std")
    aa("--phase", default=1, help="phase 1 or 2 [%(default)s]")

    f = goldstd2sec
    ap = sub.add_parser(f.func_name, help="create a secretory only gold standard")
    ap.set_defaults(func=f)
    aa = ap.add_argument
    host(aa)
    aa(
        "--tablename",
        help="name of secretory only table table [%(default)s]",
        default="gold_std_sec_only",
    )
    aa("--gold_std", help="main gold standard table [%(default)s]", default="gold_std")

    f = combine_phases
    ap = sub.add_parser(
        f.func_name,
        help="combine phase2 (secretory only) and phase1 table",
    )
    ap.set_defaults(func=f)
    aa = ap.add_argument
    host(aa)
    aa(
        "--subacon",
        dest="tablename",
        help="name of new raw subacon table [%(default)s]",
        default="subacon",
    )
    aa("--phase1", help="name of  phase1 table [%(default)s]", default="subacon_p1")
    aa("--phase2", help="name of  phase2 table [%(default)s]", default="subacon_p2")
    aa("--percent", help="percent for call [%(default)s]", default=0.5, type=float)
    aa("--merge", help="name of  location to merge [%(default)s]", default="secretory")

    f = subacon
    ap = sub.add_parser(f.func_name, help="generate consensus table")
    ap.set_defaults(func=f)
    aa = ap.add_argument
    host(aa)
    aa("--tablename", help="subacon table to create", required=True)
    aa(
        "--gold_std",
        help="gold standard table  to use [%(default)s]",
        default="gold_std",
    )
    aa("--fract", default=0.5, type=float, help="fraction to make a call [%(default)s]")
    aa("--phase", help="phase to use [%(default)s]", default="1", choices=("1", "2"))
    aa(
        "--use_exp",
        help='use experimental data (either "all" or a comma separated set of tables) [%(default)s]',
        default="all",
    )
    aa(
        "--use_pred",
        help='use predictors (either "all" or a comma separated set of tables) [%(default)s]',
        default="all",
    )

    f = merge_gold_std
    ap = sub.add_parser(
        f.func_name,
        help="update consensus table with gold standard and others [AtM..] [AtC..]",
    )
    ap.set_defaults(func=f)
    aa = ap.add_argument
    aa(
        "--gold_std",
        help="gold standard table [%(default)s]",
        required=True,
        default="gold_std",
    )
    aa("--subacon", help="subacon table", required=True)


def setup_host(host):
    # pylint: disable=global-statement
    global tair10_engine, Meta3
    if "?" not in host:
        host = host + "?use_unicode=0&charset=utf8"
    tair10_engine = create_engine(host)
    # tair10_engine.echo=True
    Meta3 = MetaData(bind=tair10_engine)
    Meta3.reflect()  # get all tables


def run():
    Args = argparse.ArgumentParser(description="subacon predictor")

    sub = Args.add_subparsers(help="sub-command help")
    cmds(sub)
    args = Args.parse_args()
    setup_host(host=args.host)
    func = args.func
    del args.func
    del args.host
    args = dict(args._get_kwargs())  # pylint: disable=protected-access
    func(**args)


if __name__ == "__main__":
    run()
